/*
** rts/event.c
**
** Viper Event Log functions
**
** Copyright (c) 1998 Peter Eriksson <pen@lysator.liu.se>
*/

#include <stdio.h>
#include <string.h>
#include <syslog.h>

#include "rts/rts.h"
#include "rts/event.h"


static double
f2c(double f)
{
    return (f - 32) / 1.8;
}


static char *
rts_str_rsi(char *cp,
	    struct rts_event_raid_set_information *rsi)
{
    if (rsi->raid_set != 255)
    {
	sprintf(cp, "RAID Set #%d (", rsi->raid_set);
	while (*cp) ++cp;
	sprintf(cp, "Level=%s, ", rts_raid2str(rsi->raid_level));
	while (*cp) ++cp;
	sprintf(cp, "Status=%s): ", rts_status2str(rsi->raid_set_status));
	while (*cp) ++cp;
    }

    if (rsi->redundancy_group != 255)
    {
	sprintf(cp, "RG #%d ", rsi->redundancy_group);
	while (*cp) ++cp;

	if (rsi->member != 255)
	{
	    sprintf(cp, "(Member #%d/", rsi->member);
	    while (*cp) ++cp;
	
	    if (rsi->number_of_members != 255)
		sprintf(cp, "%d", rsi->number_of_members);
	    else
		sprintf(cp, "???");
	    strcat(cp, ") ");
	    while (*cp) ++cp;
	}

	strcat(cp, ": ");
	while (*cp) ++cp;
    }
	
    if (rsi->raid_set_partitions != 255)
    {
	sprintf(cp, "RAID Set Partitions=%d: ", rsi->raid_set_partitions);
	while (*cp) ++cp;
    }
	
    if (rsi->number_of_redundancy_groups)
    {
	sprintf(cp, "Redundancy groups=%d: ", rsi->number_of_redundancy_groups);
	while (*cp) ++cp;
    }

    return cp;
}


static void
rts_str_rsi_v(FILE *fp,
	      struct rts_event_raid_set_information *rsi)
{
    if (rsi->raid_set != 255)
    {
	fprintf(fp, "%-20s", "RAID Set");
	if (rsi->raid_set != 255)
	    fprintf(fp, "%-30d", rsi->raid_set);
	else
	    fprintf(fp, "%-30s", "NA");
    
	fprintf(fp, "%-20s", "RAID Level");
	if (rsi->raid_level != 255)
	    fprintf(fp, "%s", rts_raid2str(rsi->raid_level));
	else
	    fprintf(fp, "%s", "NA");
	
	putc('\n', fp);

    
	fprintf(fp, "%-20s", "RAID Set Status");
	if (rsi->raid_set_status != 255)
	    fprintf(fp, "%-30s", rts_status2str(rsi->raid_set_status));
	else
	    fprintf(fp, "%-30s", "NA");
	
	fprintf(fp, "%-20s", "Redundancy Group");
	if (rsi->redundancy_group != 255)
	    fprintf(fp, "%d", rsi->redundancy_group);
	else
	    fprintf(fp, "%s", "NA");
	
	putc('\n', fp);

	fprintf(fp, "%-20s", "Logical Member");
	if (rsi->member != 255)
	    fprintf(fp, "%-30d", rsi->member);
	else
	    fprintf(fp, "%-30s", "NA");
	
	fprintf(fp, "%-20s", "Partitions");
	if (rsi->raid_set_partitions != 255)
	    fprintf(fp, "%d", rsi->raid_set_partitions);
	else
	    fprintf(fp, "%s", "NA");
	
	putc('\n', fp);
	putc('\n', fp);
    }

#if 0
    if (rsi->number_of_members != 255)
	fprintf(fp, "%d", rsi->number_of_members);
    else
	fprintf(fp, "???");
	
    if (rsi->number_of_redundancy_groups)
    {
	fprintf(fp, "Redundancy groups=%d ", rsi->number_of_redundancy_groups);
    }
#endif
}

const char *
rts_strevtype(int event_type)
{
    switch (event_type)
    {
      case RTS_UNKNOWN_EVENT:
	return "Unknown";
	break;

      case RTS_SCSI_EVENT:
	return "SCSI";
	break;

      case RTS_PORT_EVENT:
	return "Port";
	break;

      case RTS_PARITY_ERROR_EVENT:
	return "Parity Error";
	break;

      case RTS_CREATE_STARTED_EVENT:
	return "Create Started";
	break;

      case RTS_REBUILD_STARTED_EVENT:
	return "Rebuild Started";
	break;

      case RTS_CREATE_COMPLETED_EVENT:
	return "Create Completed";
	break;

      case RTS_REBUILD_COMPLETED_EVENT:
	return "Rebuld Completed";
	break;

      case RTS_DROPPED_DRIVE_EVENT:
	return "Dropped Drive";
	break;

      case RTS_TEMPERATURE_CHANGED_EVENT:
	return "Temperature Changed";
	break;

      case RTS_BAD_FAN_EVENT:
	return "Bad Fan";
	break;

      case RTS_DRIVE_SEEN_EVENT:
	return "Drive Seen";
	break;

      case RTS_DRIVE_GONE_EVENT:
	return "Drive Gone";
	break;

      case RTS_VALIDATION_CONFLICT_EVENT:
	return "Validation Conflict";
	break;

      case RTS_SCSI_HOST_EVENT:
	return "SCSI Host";
	break;

      case RTS_REASSIGN_BLOCK_EVENT:
	return "Reassign Block";
	break;

      case RTS_INTERNAL_ENVIRONMENTAL_SYSTEM_EVENT:
	return "Internal Environmental System";
	break;

      case RTS_PORT_SCSI_SYNCHRONOUS_NEGOTIATION_EVENT:
	return "Port Sync Negotiation";
	break;

      case RTS_PORT_SCSI_WIDE_DMA_NEGOTIATION_EVENT:
	return "Port Wide Negotiation";
	break;

      case RTS_HOST_SCSI_SYNCHRONOUS_NEGOTIATION_EVENT:
	return "Host Sync Negotiation";
	break;

      case RTS_HOST_SCSI_WIDE_DMA_NEGOTIATION_EVENT:
	return "Host Wide Negotiation";
	break;

      case RTS_FORCED_ERROR_WRITTEN_EVENT:
	return "Forced Error Written";
	break;

      case RTS_UNEXPECTED_PORT_INTERRUPT_EVENT:
	return "Unexpected Port Interrupt";
	break;

      case RTS_ADD_SPARE_STARTED_EVENT:
	return "Add Spare Started";
	break;

      case RTS_ADD_SPARE_COMPLETED_EVENT:
	return "Add Spare Completed";
	break;

      case RTS_SYSTEM_POWER_UP_EVENT:
	return "System Power-Up";
	break;

      case RTS_SYSTEM_SHUTDOWN_EVENT:
	return "System Shutdown";
	break;

      case RTS_REDUNDANT_CONTROLLER_TRANSMIT_MESSAGE_EVENT:
	return "Redundant Controller Transmit Message";
	break;

      case RTS_REDUNDANT_CONTROLLER_RECEIVE_MESSAGE_EVENT:
	return "Redundant Controller Receive Message";
	break;

      case RTS_REDUNDANT_CONTROLLER_STATUS_CHANGE_EVENT:
	return "Redundant Controller Status Change";
	break;

      case RTS_REDUNDANT_CONTROLLER_PARITY_ERROR_EVENT:
	return "Redundant Controller Parity Error";
	break;

      case RTS_REDUNDANT_CONTROLLER_PARITY_ERROR_ACTION_EVENT:
	return "Redundant Controller Parity Error Action";
	break;

      case RTS_FORMAT_STARTED_EVENT:
	return "Format Started";
	break;

      case RTS_FORMAT_COMPLETED_EVENT:
	return "Format Completed";
	break;

      case RTS_DRIVE_VALIDATION_EVENT:
	return "Drive Validation";
	break;
    }
    
    return "Unknown";
}

static int
dma_width_to_int(int wi)
{
    switch (wi)
    {
      case 0:
	return 8;

      case 1:
	return 16;

      case 2:
	return 32;

      default:
	return -1;
    }
}

const char *
scsi_str_error_type(int t)
{
    switch (t)
    {
      case 0:
	return "No Error";

      case 1:
	return "Received Clear Queue Message";
	
      case 2:
	return "Received Abort Message";

      case 3:
	return "Received Abort Tag Message";

      case 4:
	return "Reselection Timeout";

      case 5:
	return "Aborted Initiator/LUN Tasks";

      case 6:
	return "Aborted Task";

      case 7:
	return "Busy";

      case 8:
	return "Reserved (#8)";

      case 9:
	return "Queue Full";

      case 10:
	return "Busy/Overlapped Command";

      case 11:
	return "Device Reset";

      case 12:
	return "Reselected by Another SCSI Device";

      case 13:
	return "Connection Timed Out by Host Port";

      case 14:
	return "Delay Mode";

      default:
	return "Unknown";
    }
}



const char *
rts_str_event(struct rts_event_information *ep)
{
    static char buf[2048];
    char *cp;
    int i;
    

    cp = buf;
    
    sprintf(cp, "[%d] ", rts_w2i(ep->sequence_number));
    while (*cp)	++cp;
    
#if 0    
    bt = rts_mktime(&ep->time_stamp);
    cftime(cp, "%D %T: ", &bt);
    while (*cp)	++cp;
#endif
    
    sprintf(cp, "%s Event: ", rts_strevtype(ep->event_type));
    while (*cp) ++cp;
    
    switch (ep->event_type)
    {
      case RTS_UNKNOWN_EVENT:
	sprintf(cp, "Type=%d", ep->data.unknown.unknown_event_value);
	break;

      case RTS_SCSI_EVENT:
	sprintf(cp, "Target=c%dt%d: ",
		ep->data.scsi.channel,
		ep->data.scsi.scsi_id);
	while (*cp) ++cp;
	cp = rts_str_rsi(cp, &ep->data.scsi.raid_set_information);
	break;

      case RTS_PORT_EVENT:
#if 0
	cp = rts_str_rsi(cp, &ep->data.port.raid_set_information);
#endif

	sprintf(cp, "Target=c%dt%d: ",
		ep->data.port.channel,
		ep->data.port.scsi_id);

	while (*cp) ++cp;

	sprintf(cp, "Status=%s",
		rts_strpstatus(ep->data.port.port_status));

	break;

      case RTS_PARITY_ERROR_EVENT:
	sprintf(cp, "Memory address %08X",
		rts_q2i(ep->data.parity_error.memory_address));
	break;

      case RTS_CREATE_STARTED_EVENT:
	cp = rts_str_rsi(cp, &ep->data.create_started.raid_set_information);
	sprintf(cp, ", Members:");
	while (*cp) ++cp;
	
	for (i = 0; i < ep->data.create_started.raid_set_information.number_of_members; i++)
	{
	    sprintf(cp, " c%dt%d",
		    ep->data.create_started.member[i].channel,
		    ep->data.create_started.member[i].scsi_id);
	    while (*cp) ++cp;
	}
	    
	break;

      case RTS_REBUILD_STARTED_EVENT:
	cp = rts_str_rsi(cp, &ep->data.rebuild_started.raid_set_information);
	sprintf(cp, ", Member: c%dt%d",
		ep->data.rebuild_started.channel,
		ep->data.rebuild_started.scsi_id);
	break;

      case RTS_CREATE_COMPLETED_EVENT:
	cp = rts_str_rsi(cp, &ep->data.create_completed.raid_set_information);
	sprintf(cp, ", Status: %s",
		ep->data.create_completed.status ? "Failed" : "Success");
		
	if (ep->data.create_completed.status != 0)
	{
	    sprintf(cp, ", Member: c%dt%d",
		    ep->data.create_completed.channel,
		    ep->data.create_completed.scsi_id);
	}
	break;

      case RTS_REBUILD_COMPLETED_EVENT:
	cp = rts_str_rsi(cp, &ep->data.rebuild_completed.raid_set_information);
	sprintf(cp, ", Status: %s",
		ep->data.rebuild_completed.status ? "Failed" : "Success");
		
	if (ep->data.rebuild_completed.status != 0)
	{
	    sprintf(cp, ", Member: c%dt%d",
		    ep->data.rebuild_completed.channel,
		    ep->data.rebuild_completed.scsi_id);
	}
	break;

      case RTS_DROPPED_DRIVE_EVENT:
	cp = rts_str_rsi(cp, &ep->data.dropped_drive.raid_set_information);
	sprintf(cp, ", Reason: %s",
		ep->data.dropped_drive.reason ?
		"User Initiated" : "Drive Failure");;
	break;

      case RTS_TEMPERATURE_CHANGED_EVENT:
	sprintf(cp, "Temperature: %dF (%.1fC)",
		ep->data.temperature_changed.temperature,
		f2c(ep->data.temperature_changed.temperature));
	break;

      case RTS_BAD_FAN_EVENT:
	sprintf(cp, "Fan number: %d", ep->data.bad_fan.fan);
	break;

      case RTS_DRIVE_SEEN_EVENT:
	sprintf(cp, "Drive: c%dt%d",
		ep->data.drive_seen.channel,
		ep->data.drive_seen.scsi_id);
	break;

      case RTS_DRIVE_GONE_EVENT:
	sprintf(cp, "Drive: c%dt%d",
		ep->data.drive_gone.channel,
		ep->data.drive_gone.scsi_id);
	break;

      case RTS_VALIDATION_CONFLICT_EVENT:
	cp = rts_str_rsi(cp, &ep->data.validation_conflict.raid_set_information);
	/* XXX: More here */
	break;

      case RTS_SCSI_HOST_EVENT:
	strcpy(cp, scsi_str_error_type(ep->data.scsi_host.error_type));
	
	/* XXX: More here */
	break;

      case RTS_REASSIGN_BLOCK_EVENT:
	cp = rts_str_rsi(cp, &ep->data.reassign_block.raid_set_information);
	sprintf(cp, ", Member: c%dt%d, Status: %s",
		ep->data.reassign_block.channel,
		ep->data.reassign_block.scsi_id,
		ep->data.reassign_block.status == 0 ?
		"Success" : "Failure");

	/* XXX: More here */
	break;

      case RTS_INTERNAL_ENVIRONMENTAL_SYSTEM_EVENT:
	switch (ep->data.internal_env_system.actual_device)
	{
	  case 0:
	    strcpy(cp, "External Temperature");
	    break;
	    
	  case 1:
	    strcpy(cp, "Viper Board Temperature");
	    break;
	    
	  case 2:
	    strcpy(cp, "Termination Power");
	    break;
	    
	  case 3:
	    strcpy(cp, "Cache DRAM #2 Power");
	    break;
	    
	  case 4:
	    strcpy(cp, "System 5V Power");
	    break;
	    
	  case 5:
	    strcpy(cp, "Cache DRAM #1 Power");
	    break;
	    
	  case 6:
	    strcpy(cp, "System 12V Power");
	    break;
	    
	  case 7:
	    strcpy(cp, "Battery");
	    break;
	    
	  case 8:
	    strcpy(cp, "UPS");
	    break;
	}

	while (*cp) ++cp;
	strcpy(cp, ", ");
	while (*cp) ++cp;
	
	switch (ep->data.internal_env_system.type_of_device)
	{
	  case 0:
	    sprintf(cp, "Temperature: %dF (%.1fC)\n",
		    rts_w2i(ep->data.internal_env_system.value),
		    f2c(rts_w2i(ep->data.internal_env_system.value)));
	    break;
	    
	  case 1:
	    sprintf(cp, "Voltage: %dV\n",
		    rts_w2i(ep->data.internal_env_system.value));
	    break;
	    
	}

	while (*cp) ++cp;
	strcpy(cp, ", Status: ");
	while (*cp) ++cp;
	
	switch (ep->data.internal_env_system.device_status)
	{
	  case 0:
	    strcpy(cp, "Not Installed");
	    break;

	  case 1:
	    strcpy(cp, "Normal");
	    break;
	    
	  case 2:
	    strcpy(cp, "Below Normal");
	    break;
	    
	  case 3:
	    strcpy(cp, "Above Normal");
	    break;
	    
	  case 4:
	    strcpy(cp, "Too Low");
	    break;
	    
	  case 5:
	    strcpy(cp, "Too High");
	    break;
	    
	  case 6:
	    strcpy(cp, "Back to Normal");
	    break;
	    
	  case 7:
	    strcpy(cp, "AC Power Lost");
	    break;
	    
	  case 8:
	    strcpy(cp, "2 Minute Warning");
	    break;
	    
	  default:
	    sprintf(cp, "Unknown (%d)",
		    ep->data.internal_env_system.device_status);
	}
	
	break;

      case RTS_PORT_SCSI_SYNCHRONOUS_NEGOTIATION_EVENT:
	sprintf(cp, "Member=c%dt%d: DMA Tx Time=%d ns: Tx Off: %d bytes",
		ep->data.port_sync_negotiation.channel,
		ep->data.port_sync_negotiation.scsi_id,

		rts_w2i(ep->data.port_sync_negotiation.received_sync_time_period),
		ep->data.port_sync_negotiation.received_sync_offset);
	break;

      case RTS_PORT_SCSI_WIDE_DMA_NEGOTIATION_EVENT:
	sprintf(cp, "Member=c%dt%d: Width=%d bits",
		ep->data.port_wide_negotiation.channel,
		ep->data.port_wide_negotiation.scsi_id,
		dma_width_to_int(ep->data.port_wide_negotiation.received_dma_width));
	break;

      case RTS_HOST_SCSI_SYNCHRONOUS_NEGOTIATION_EVENT:
	sprintf(cp, "Initiator=%d: LUM=%d: DMA Tx Time=%d ns: Tx Off: %d bytes",
		ep->data.host_sync_negotiation.host_initiator_id,
		ep->data.host_sync_negotiation.host_lun,

		rts_w2i(ep->data.host_sync_negotiation.returned_sync_time_period),
		ep->data.host_sync_negotiation.returned_sync_offset);
	break;

      case RTS_HOST_SCSI_WIDE_DMA_NEGOTIATION_EVENT:
	sprintf(cp, "Inititator=%d, LUN=%d: Width=%d bits",
		ep->data.host_wide_negotiation.host_initiator_id,
		ep->data.host_wide_negotiation.host_lun,
		dma_width_to_int(ep->data.host_wide_negotiation.returned_dma_width));
	break;

      case RTS_FORCED_ERROR_WRITTEN_EVENT:
	break;

      case RTS_UNEXPECTED_PORT_INTERRUPT_EVENT:
	break;

      case RTS_ADD_SPARE_STARTED_EVENT:
	break;

      case RTS_ADD_SPARE_COMPLETED_EVENT:
	break;

      case RTS_SYSTEM_POWER_UP_EVENT:
	break;

      case RTS_SYSTEM_SHUTDOWN_EVENT:
	break;

      case RTS_REDUNDANT_CONTROLLER_TRANSMIT_MESSAGE_EVENT:
	break;

      case RTS_REDUNDANT_CONTROLLER_RECEIVE_MESSAGE_EVENT:
	break;

      case RTS_REDUNDANT_CONTROLLER_STATUS_CHANGE_EVENT:
	break;

      case RTS_REDUNDANT_CONTROLLER_PARITY_ERROR_EVENT:
	break;

      case RTS_REDUNDANT_CONTROLLER_PARITY_ERROR_ACTION_EVENT:
	break;

      case RTS_FORMAT_STARTED_EVENT:
	break;

      case RTS_FORMAT_COMPLETED_EVENT:
	break;

      case RTS_DRIVE_VALIDATION_EVENT:
	break;

      default:
	sprintf(cp, "Type: %d", ep->event_type);
    }
    
    return buf;
}



void
rts_print_event(FILE *fp,
		struct rts_event_information *ep)
{
    time_t bt;
    char buf[1024];
    int i;
    
    
    bt = rts_mktime(&ep->time_stamp);
    
    cftime(buf, "%D", &bt);
    fprintf(fp, "%-20s%-30d%-20s%s\n", 
	    "Sequence Number",
	    rts_w2i(ep->sequence_number),
	    "Date",
	    buf);
    
    cftime(buf, "%T", &bt);
    fprintf(fp, "%-20s%-30s%-20s%s\n\n",
	    "Recorded Event",
	    rts_strevtype(ep->event_type),
	    "Time",
	    buf);
    
    switch (ep->event_type)
    {
      case RTS_UNKNOWN_EVENT:
	break;

      case RTS_SCSI_EVENT:
	rts_str_rsi_v(fp, &ep->data.scsi.raid_set_information);
	fprintf(fp, "%-20s%d\n", "I/O Channel",
		ep->data.scsi.channel);
	fprintf(fp, "%-20s%d\n", "SCSI ID",
		ep->data.scsi.scsi_id);
	fprintf(fp, "%-20s",
		"SCSI Command");
	rts_print_hex(fp,
	       ep->data.scsi.scsi_command,
	       sizeof(ep->data.scsi.scsi_command));
	putc('\n', fp);
	fprintf(fp, "%-20s",
		"SCSI Sense Data");
	rts_print_hex(fp,
	       ep->data.scsi.request_sense_data,
	       sizeof(ep->data.scsi.request_sense_data));
	putc('\n', fp);
	fprintf(fp, "%-20s",
		"SCSI Phases");
	rts_print_hex(fp,
	       ep->data.scsi.phase_history,
	       sizeof(ep->data.scsi.phase_history));
	putc('\n', fp);
	fprintf(fp, "%-20s%d\n",
		"Retry Attempt",
		ep->data.scsi.retry_number);
	fprintf(fp, "%-20s%08X\n",
		"SCDRP",
		rts_q2i(ep->data.scsi.scdrp));
	fprintf(fp, "%-20s%08X\n",
		"CDRP",
		rts_q2i(ep->data.scsi.cdrp));
	putc('\n', fp);
	break;

      case RTS_PORT_EVENT:
	rts_str_rsi_v(fp, &ep->data.port.raid_set_information);
	fprintf(fp, "%-20s%d\n", "I/O Channel",
		ep->data.port.channel);
	fprintf(fp, "%-20s%d\n", "SCSI ID",
		ep->data.port.scsi_id);
	fprintf(fp, "%-20s%s\n",
		"Port Status",
		rts_strpstatus(ep->data.port.port_status));

	fprintf(fp, "%-20s",
		"SCSI Command");
	rts_print_hex(fp,
	       ep->data.port.scsi_command,
	       sizeof(ep->data.port.scsi_command));
	putc('\n', fp);
	fprintf(fp, "%-20s",
		"SCSI Phases");
	rts_print_hex(fp,
	       ep->data.port.phase_history,
	       sizeof(ep->data.port.phase_history));
	putc('\n', fp);
	fprintf(fp, "%-20s%d\n",
		"Retry Attempt",
		ep->data.port.retry_number);
	fprintf(fp, "%-20s%08X\n",
		"SCDRP",
		rts_q2i(ep->data.port.scdrp));
	fprintf(fp, "%-20s%08X\n",
		"CDRP",
		rts_q2i(ep->data.port.cdrp));
	putc('\n', fp);
	break;

      case RTS_PARITY_ERROR_EVENT:
	fprintf(fp, "%-20s%08X\n\n",
		"Memory Address",
		rts_q2i(ep->data.parity_error.memory_address));
	break;
	
      case RTS_CREATE_STARTED_EVENT:
	rts_str_rsi_v(fp, &ep->data.create_started.raid_set_information);
	fprintf(fp, "%-20s",
		"RAID Set Members");
	for (i = 0; i < ep->data.create_started.raid_set_information.number_of_members; i++)
	    fprintf(fp, " c%dt%d",
		    ep->data.create_started.member[i].channel,
		    ep->data.create_started.member[i].scsi_id);
	putc('\n', fp);
	putc('\n', fp);
	break;

      case RTS_REBUILD_STARTED_EVENT:
	rts_str_rsi_v(fp, &ep->data.rebuild_started.raid_set_information);
	fprintf(fp, "%-20s%d\n",
		"Channel",
		ep->data.rebuild_started.channel);
	fprintf(fp, "%-20s%d\n",
		"SCSI ID",
		ep->data.rebuild_started.scsi_id);
	putc('\n', fp);
	break;

      case RTS_CREATE_COMPLETED_EVENT:
	rts_str_rsi_v(fp, &ep->data.create_completed.raid_set_information);
	fprintf(fp, "%-20s%s\n",
		"Status",
		ep->data.create_completed.status ? "Failed" : "Success");
	
	if (ep->data.create_completed.status != 0)
	{
	    fprintf(fp, "%-20s%d\n",
		    "Channel",
		    ep->data.create_completed.channel);
	    fprintf(fp, "%-20s%d\n",
		    "SCSI ID",
		    ep->data.create_completed.scsi_id);
	}
	putc('\n', fp);
	break;

      case RTS_REBUILD_COMPLETED_EVENT:
	rts_str_rsi_v(fp, &ep->data.rebuild_completed.raid_set_information);
	fprintf(fp, "%-20s%s\n",
		"Status",
		ep->data.rebuild_completed.status ? "Failed" : "Success");
	
	if (ep->data.rebuild_completed.status != 0)
	{
	    fprintf(fp, "%-20s%d\n",
		    "Channel",
		    ep->data.rebuild_completed.channel);
	    fprintf(fp, "%-20s%d\n",
		    "SCSI ID",
		    ep->data.rebuild_completed.scsi_id);
	}
	putc('\n', fp);
	break;

      case RTS_DROPPED_DRIVE_EVENT:
	rts_str_rsi_v(fp, &ep->data.dropped_drive.raid_set_information);
	fprintf(fp, "%-20s%s\n",
		"Reason",
		ep->data.dropped_drive.reason ? "User Initiated" : "Drive Failure");
	
	fprintf(fp, "%-20s%d\n",
		"Channel",
		ep->data.dropped_drive.channel);
	fprintf(fp, "%-20s%d\n\n",
		"SCSI ID",
		ep->data.dropped_drive.scsi_id);
	break;

      case RTS_TEMPERATURE_CHANGED_EVENT:
	fprintf(fp, "%-20s%dF (%.1fC)\n\n",
		"Temperature",
		ep->data.temperature_changed.temperature,
		f2c(ep->data.temperature_changed.temperature));
	break;

      case RTS_BAD_FAN_EVENT:
	fprintf(fp, "%-20s%d\n\n",
		"Fan Number",
		ep->data.bad_fan.fan);
	break;

      case RTS_DRIVE_SEEN_EVENT:
	fprintf(fp, "%-20s%d\n",
		"Channel",
		ep->data.drive_seen.channel);
	fprintf(fp, "%-20s%d\n\n",
		"SCSI ID",
		ep->data.drive_seen.scsi_id);
	break;

      case RTS_DRIVE_GONE_EVENT:
	fprintf(fp, "%-20s%d\n",
		"Channel",
		ep->data.drive_gone.channel);
	fprintf(fp, "%-20s%d\n\n",
		"SCSI ID",
		ep->data.drive_gone.scsi_id);
	break;

      case RTS_VALIDATION_CONFLICT_EVENT:
	rts_str_rsi_v(fp, &ep->data.validation_conflict.raid_set_information);
	fprintf(fp, "\n<SORRY - THIS CODE HAS YET TO BE WRITTEN>\n\n");
	break;

      case RTS_SCSI_HOST_EVENT:
	fprintf(fp, "%-20s%08X\n",
		"CDRP",
		rts_q2i(ep->data.scsi_host.cdrp));
	fprintf(fp, "%-20s%d\n",
		"Target ID",
		ep->data.scsi_host.target_id);
	fprintf(fp, "%-20s%d\n",
		"LUN",
		ep->data.scsi_host.lun);
	fprintf(fp, "%-20s%d\n",
		"Initator ID",
		ep->data.scsi_host.initiator_id);
	fprintf(fp, "%-20s%02X\n",
		"HTCB Flags",
		ep->data.scsi_host.htcb_flags);
	fprintf(fp, "%-20s%04X\n",
		"CDRP Host Flags",
		rts_w2i(ep->data.scsi_host.cdrp_host_flags));
	fprintf(fp, "%-20s%04X\n",
		"CDRP Flags",
		rts_w2i(ep->data.scsi_host.cdrp_flags));
	fprintf(fp, "%-20s%s", "Error Type",
		scsi_str_error_type(ep->data.scsi_host.error_type));
	
	putc('\n', fp);
	fprintf(fp, "%-20s", "Tag Type");
	switch (ep->data.scsi_host.tag_type)
	{
	  case 0x00:
	    fputs("No Tag", fp);
	    break;

	  case 0x20:
	    fputs("Simple Tag", fp);
	    break;

	  case 0x21:
	    fputs("Head Tag", fp);
	    break;

	  case 0x22:
	    fputs("Ordered Tag", fp);
	    break;

	  default:
	    fputs("???", fp);
	}
	putc('\n', fp);
	putc('\n', fp);
	
	fprintf(fp, "%-50s%02X\n",
		"SCSI Chip Status Register",
		ep->data.scsi_host.scsi_chip_status_register);
	fprintf(fp, "%-50s%02X\n",
		"SCSI Chip Interrupt Register",
		ep->data.scsi_host.scsi_chip_interrupt_register);
	fprintf(fp, "%-50s%d\n",
		"SCSI Status",
		ep->data.scsi_host.scsi_status);
	putc('\n', fp);
	fprintf(fp, "%-20s",
		"SCSI Command");
	rts_print_hex(fp,
	       ep->data.scsi_host.scsi_command,
	       sizeof(ep->data.scsi_host.scsi_command));
	putc('\n', fp);
	fprintf(fp, "%-20s",
		"SCSI Sense Data");
	rts_print_hex(fp,
	       ep->data.scsi_host.request_sense,
	       sizeof(ep->data.scsi_host.request_sense));
	putc('\n', fp);
	putc('\n', fp);
	break;

      case RTS_REASSIGN_BLOCK_EVENT:
	rts_str_rsi_v(fp, &ep->data.reassign_block.raid_set_information);
	
	fprintf(fp, "%-20s%d\n", "I/O Channel",
		ep->data.reassign_block.channel);
	fprintf(fp, "%-20s%d\n", "SCSI ID",
		ep->data.reassign_block.scsi_id);
	fprintf(fp, "%-20s%s\n",
		"Status",
		ep->data.reassign_block.status == 0 ?
		"Success" : "Failure");

	fprintf(fp, "%-20s%08X\n",
		"SCDRP",
		rts_q2i(ep->data.reassign_block.scdrp));
	fprintf(fp, "%-20s%08X\n",
		"CDRP",
		rts_q2i(ep->data.reassign_block.cdrp));

	if (rts_q2i(ep->data.reassign_block.reassigned_vbn) != 0xFFFFFFFF)
	    fprintf(fp, "%-40s%08X\n",
		    "Reassigned Virtual Block Number",
		    rts_q2i(ep->data.reassign_block.reassigned_vbn));
	fprintf(fp, "%-40s%08X\n",
		"Reassigned Physical Block Number",
		rts_q2i(ep->data.reassign_block.reassigned_pbn));

	putc('\n', fp);
	break;

      case RTS_INTERNAL_ENVIRONMENTAL_SYSTEM_EVENT:
	fprintf(fp, "%-20s", "Device");
	switch (ep->data.internal_env_system.actual_device)
	{
	  case 0:
	    fputs("External Temperature\n", fp);
	    break;
	    
	  case 1:
	    fputs("Viper Board Temperature\n", fp);
	    break;
	    
	  case 2:
	    fputs("Termination Power\n", fp);
	    break;
	    
	  case 3:
	    fputs("Cache DRAM #2 Power\n", fp);
	    break;
	    
	  case 4:
	    fputs("System 5V Power\n", fp);
	    break;
	    
	  case 5:
	    fputs("Cache DRAM #1 Power\n", fp);
	    break;
	    
	  case 6:
	    fputs("System 12V Power\n", fp);
	    break;
	    
	  case 7:
	    fputs("Battery\n", fp);
	    break;
	    
	  case 8:
	    fputs("UPS\n", fp);
	    break;
	}
	
	switch (ep->data.internal_env_system.type_of_device)
	{
	  case 0:
	    fprintf(fp, "%-20s%dF (%.1fC)\n",
		    "Temperature",
		    rts_w2i(ep->data.internal_env_system.value),
		    f2c(rts_w2i(ep->data.internal_env_system.value)));
	    break;
	    
	  case 1:
	    fprintf(fp, "%-20s%dV\n",
		    "Voltage",
		    rts_w2i(ep->data.internal_env_system.value));
	    break;
	    
	}
	
	fprintf(fp, "%-20s", "Status");

	switch (ep->data.internal_env_system.device_status)
	{
	  case 0:
	    fputs("Not Installed\n", fp);
	    break;

	  case 1:
	    fputs("Normal\n", fp);
	    break;
	    
	  case 2:
	    fputs("Below Normal\n", fp);
	    break;
	    
	  case 3:
	    fputs("Above Normal\n", fp);
	    break;
	    
	  case 4:
	    fputs("Too Low\n", fp);
	    break;
	    
	  case 5:
	    fputs("Too High\n", fp);
	    break;
	    
	  case 6:
	    fputs("Back to Normal\n", fp);
	    break;
	    
	  case 7:
	    fputs("AC Power Lost\n", fp);
	    break;
	    
	  case 8:
	    fputs("2 Minute Warning\n", fp);
	    break;
	    
	  default:
	    fprintf(fp, "Unknown (%d)\n",
		    ep->data.internal_env_system.device_status);
	}
	putc('\n', fp);
	break;

      case RTS_PORT_SCSI_SYNCHRONOUS_NEGOTIATION_EVENT:
	fprintf(fp, "%-50s%d\n", "I/O Channel",
		ep->data.port_sync_negotiation.channel);
	fprintf(fp, "%-50s%d\n", "SCSI ID",
		ep->data.port_sync_negotiation.scsi_id);

	fprintf(fp, "%-50s%d ns\n", "Requested Sync DMA Transfer Time",
		rts_w2i(ep->data.port_sync_negotiation.requested_sync_time_period));
	fprintf(fp, "%-50s%d ns\n", "Received Sync DMA Transfer Time",
		rts_w2i(ep->data.port_sync_negotiation.received_sync_time_period));
	fprintf(fp, "%-50s%d bytes\n", "Requested Sync DMA Transfer Offset",
		ep->data.port_sync_negotiation.requested_sync_offset);
	fprintf(fp, "%-50s%d bytes\n", "Received Sync DMA Transfer Offset",
		ep->data.port_sync_negotiation.received_sync_offset);
	putc('\n', fp);
	break;

      case RTS_PORT_SCSI_WIDE_DMA_NEGOTIATION_EVENT:
	fprintf(fp, "%-50s%d\n", "I/O Channel",
		ep->data.port_wide_negotiation.channel);
	fprintf(fp, "%-50s%d\n", "SCSI ID",
		ep->data.port_wide_negotiation.scsi_id);

	
	fprintf(fp, "%-50s%d bits\n", "Requested DMA Width",
		dma_width_to_int(ep->data.port_wide_negotiation.requested_dma_width));
	
	fprintf(fp, "%-50s%d bits\n\n", "Received DMA Width",
		dma_width_to_int(ep->data.port_wide_negotiation.received_dma_width));

	break;

      case RTS_HOST_SCSI_SYNCHRONOUS_NEGOTIATION_EVENT:
	fprintf(fp, "%-50s%d\n", "Initiator ID",
		ep->data.host_sync_negotiation.host_initiator_id);
	fprintf(fp, "%-50s%d\n", "LUN",
		ep->data.host_sync_negotiation.host_lun);
	
	fprintf(fp, "%-50s%d ns\n", "Requested Sync DMA Transfer Time",
		rts_w2i(ep->data.host_sync_negotiation.requested_sync_time_period));
	fprintf(fp, "%-50s%d ns\n", "Returned Sync DMA Transfer Time",
		rts_w2i(ep->data.host_sync_negotiation.returned_sync_time_period));
	fprintf(fp, "%-50s%d bytes\n", "Requested Sync DMA Offset",
		ep->data.host_sync_negotiation.requested_sync_offset);
	fprintf(fp, "%-50s%d bytes\n", "Returned Sync DMA Offset",
		ep->data.host_sync_negotiation.returned_sync_offset);
	putc('\n', fp);
	break;

      case RTS_HOST_SCSI_WIDE_DMA_NEGOTIATION_EVENT:
	fprintf(fp, "%-50s%d\n", "Initiator ID",
		ep->data.host_wide_negotiation.host_initiator_id);
	fprintf(fp, "%-50s%d\n", "LUN",
		ep->data.host_wide_negotiation.host_lun);
	
	fprintf(fp, "%-50s%d bits\n", "Requested DMA Width",
		dma_width_to_int(ep->data.host_wide_negotiation.requested_dma_width));
	    
	fprintf(fp, "%-50s%d bits\n\n", "Returned DMA Width",
		dma_width_to_int(ep->data.host_wide_negotiation.returned_dma_width));
	break;

      case RTS_FORCED_ERROR_WRITTEN_EVENT:
	rts_str_rsi_v(fp, &ep->data.forced_error_written.raid_set_information);
	
	fprintf(fp, "%-20s%d\n", "I/O Channel",
		ep->data.forced_error_written.channel);
	fprintf(fp, "%-20s%d\n", "SCSI ID",
		ep->data.forced_error_written.scsi_id);
	fprintf(fp, "%-20s%s\n",
		"Status",
		ep->data.forced_error_written.status ?
		"Success" : "Failure");

	fprintf(fp, "%-20s%08X\n",
		"SCDRP",
		rts_q2i(ep->data.forced_error_written.scdrp));
	fprintf(fp, "%-20s%08X\n",
		"CDRP",
		rts_q2i(ep->data.forced_error_written.cdrp));

	if (rts_q2i(ep->data.forced_error_written.forced_error_vbn) != 0xFFFFFFFF)
	    fprintf(fp, "%-40s%08X\n",
		    "Reassigned Virtual Block Number",
		    rts_q2i(ep->data.forced_error_written.forced_error_vbn));
	fprintf(fp, "%-40s%08X\n",
		"Reassigned Physical Block Number",
		rts_q2i(ep->data.forced_error_written.forced_error_pbn));

	putc('\n', fp);
	break;

      case RTS_UNEXPECTED_PORT_INTERRUPT_EVENT:
	fprintf(fp, "\n<SORRY - THIS CODE HAS YET TO BE WRITTEN>\n\n");
	break;

      case RTS_ADD_SPARE_STARTED_EVENT:
	fprintf(fp, "%-20s%d\n", "I/O Channel",
		ep->data.add_spare_started.channel);
	fprintf(fp, "%-20s%d\n", "SCSI ID",
		ep->data.add_spare_started.scsi_id);
	fprintf(fp, "%-20s", "Type");
	switch (ep->data.add_spare_started.type)
	{
	  case 12:
	    fputs("Hot Spare\n", fp);
	    break;
	  case 13:
	    fputs("Warm Spare\n", fp);
	    break;
	  case 15:
	    fputs("No Spare (Unknown)\n", fp);
	    break;
	}
	putc('\n', fp);
	break;

      case RTS_ADD_SPARE_COMPLETED_EVENT:
	rts_str_rsi_v(fp, &ep->data.add_spare_completed.raid_set_information);
	
	fprintf(fp, "%-20s%d\n", "I/O Channel",
		ep->data.add_spare_completed.channel);
	fprintf(fp, "%-20s%d\n", "SCSI ID",
		ep->data.add_spare_completed.scsi_id);
	fprintf(fp, "%-20s%s\n", "Status",
		ep->data.add_spare_completed.status ?
		"Failure" : "Success");
	
	putc('\n', fp);
	break;

      case RTS_SYSTEM_POWER_UP_EVENT:
	/* We could print the duplicated time stamp, but why bother */
	break;

      case RTS_SYSTEM_SHUTDOWN_EVENT:
	fprintf(fp, "%-20s", "Reason");
	switch (ep->data.system_shutdown.reason)
	{
	  case 1:
	    fputs("System Shutdown\n", fp);
	    break;
	  case 2:
	    fputs("System Restart\n", fp);
	}
	putc('\n', fp);
	break;

      case RTS_REDUNDANT_CONTROLLER_TRANSMIT_MESSAGE_EVENT:
	fprintf(fp, "\n<SORRY - THIS CODE HAS YET TO BE WRITTEN>\n\n");
	break;

      case RTS_REDUNDANT_CONTROLLER_RECEIVE_MESSAGE_EVENT:
	fprintf(fp, "\n<SORRY - THIS CODE HAS YET TO BE WRITTEN>\n\n");
	break;

      case RTS_REDUNDANT_CONTROLLER_STATUS_CHANGE_EVENT:
	fprintf(fp, "\n<SORRY - THIS CODE HAS YET TO BE WRITTEN>\n\n");
	break;

      case RTS_REDUNDANT_CONTROLLER_PARITY_ERROR_EVENT:
	fprintf(fp, "\n<SORRY - THIS CODE HAS YET TO BE WRITTEN>\n\n");
	break;

      case RTS_REDUNDANT_CONTROLLER_PARITY_ERROR_ACTION_EVENT:
	fprintf(fp, "\n<SORRY - THIS CODE HAS YET TO BE WRITTEN>\n\n");
	break;

      case RTS_FORMAT_STARTED_EVENT:
	fprintf(fp, "%-20s%d\n", "I/O Channel",
		ep->data.format_started.channel);
	fprintf(fp, "%-20s%d\n", "SCSI ID",
		ep->data.format_started.scsi_id);
	break;

      case RTS_FORMAT_COMPLETED_EVENT:
	fprintf(fp, "%-20s%d\n", "I/O Channel",
		ep->data.format_completed.channel);
	fprintf(fp, "%-20s%d\n", "SCSI ID",
		ep->data.format_completed.scsi_id);
	break;

      case RTS_DRIVE_VALIDATION_EVENT:
	rts_str_rsi_v(fp, &ep->data.drive_validation.raid_set_information);
	
	fprintf(fp, "%-20s%d\n", "I/O Channel",
		ep->data.drive_validation.channel);
	fprintf(fp, "%-20s%d\n", "SCSI ID",
		ep->data.drive_validation.scsi_id);
	fprintf(fp, "%-20s", "Drive Status");
	switch (ep->data.drive_validation.drive_status)
	{
	  case 0:
	    fputs("Empty", fp);
	    break;

	  case 5:
	    fputs("Creating", fp);
	    break;

	  case 10:
	    fputs("Online", fp);
	    break;

	  case 11:
	    fputs("Rebuilding", fp);
	    break;

	  case 12:
	    fputs("Hot Spare", fp);
	    break;

	  case 13:
	    fputs("Warm Spare", fp);
	    break;

	  case 14:
	    fputs("Adding Spare", fp);
	    break;

	  case 15:
	    fputs("Status Unknown", fp);
	    break;

	  case 16:
	    fputs("Formatting", fp);
	}
	putc('\n', fp);
	fprintf(fp, "%-20s", "Validation Status");
	switch (ep->data.drive_validation.validation_status)
	{
	  case 0:
	    fputs("Success", fp);
	    break;

	  case 1:
	    fputs("Failure - Drive Changed", fp);
	    break;

	  case 2:
	    fputs("Failure - Drive Error", fp);
	    break;

	  case 3:
	    fputs("Failure - Drive Not Validable", fp);
	    break;
	}
	putc('\n', fp);
	putc('\n', fp);
	break;

      default:
	break;
    }

    putc('\n', fp);
}


int
rts_event2severity(struct rts_event_information *ep)
{
    switch (ep->event_type)
    {
      case RTS_UNKNOWN_EVENT:
	return LOG_ERR;

      case RTS_SCSI_EVENT:
	return LOG_WARNING;

      case RTS_PORT_EVENT:
	return LOG_WARNING;

      case RTS_PARITY_ERROR_EVENT:
	return LOG_ERR;
	
      case RTS_CREATE_STARTED_EVENT:
	return LOG_INFO;
	
      case RTS_REBUILD_STARTED_EVENT:
	return LOG_INFO;
	
      case RTS_CREATE_COMPLETED_EVENT:
	return LOG_INFO;
	
      case RTS_REBUILD_COMPLETED_EVENT:
	return LOG_INFO;
	
      case RTS_DROPPED_DRIVE_EVENT:
	return LOG_CRIT;
	
      case RTS_TEMPERATURE_CHANGED_EVENT:
	return LOG_NOTICE;
	
      case RTS_BAD_FAN_EVENT:
	return LOG_ERR;
	
      case RTS_DRIVE_SEEN_EVENT:
	return LOG_NOTICE;
	
      case RTS_DRIVE_GONE_EVENT:
	return LOG_ERR;
	
      case RTS_VALIDATION_CONFLICT_EVENT:
	return LOG_WARNING;
	
      case RTS_SCSI_HOST_EVENT:
	return LOG_NOTICE;
	
      case RTS_REASSIGN_BLOCK_EVENT:
	return LOG_NOTICE;
	
      case RTS_INTERNAL_ENVIRONMENTAL_SYSTEM_EVENT:
	return LOG_NOTICE;
	
      case RTS_PORT_SCSI_SYNCHRONOUS_NEGOTIATION_EVENT:
	return LOG_INFO;
	
      case RTS_PORT_SCSI_WIDE_DMA_NEGOTIATION_EVENT:
	return LOG_INFO;
	
      case RTS_HOST_SCSI_SYNCHRONOUS_NEGOTIATION_EVENT:
	return LOG_INFO;
	
      case RTS_HOST_SCSI_WIDE_DMA_NEGOTIATION_EVENT:
	return LOG_INFO;
	
      case RTS_FORCED_ERROR_WRITTEN_EVENT:
	return LOG_WARNING;
	
      case RTS_UNEXPECTED_PORT_INTERRUPT_EVENT:
	return LOG_WARNING;
	
      case RTS_ADD_SPARE_STARTED_EVENT:
	return LOG_INFO;
	
      case RTS_ADD_SPARE_COMPLETED_EVENT:
	return LOG_INFO;
	
      case RTS_SYSTEM_POWER_UP_EVENT:
	return LOG_INFO;
	
      case RTS_SYSTEM_SHUTDOWN_EVENT:
	return LOG_CRIT;
	
      case RTS_REDUNDANT_CONTROLLER_TRANSMIT_MESSAGE_EVENT:
	return LOG_INFO;
	
      case RTS_REDUNDANT_CONTROLLER_RECEIVE_MESSAGE_EVENT:
	return LOG_INFO;
	
      case RTS_REDUNDANT_CONTROLLER_STATUS_CHANGE_EVENT:
	return LOG_NOTICE;
	
      case RTS_REDUNDANT_CONTROLLER_PARITY_ERROR_EVENT:
	return LOG_ERR;
	
      case RTS_REDUNDANT_CONTROLLER_PARITY_ERROR_ACTION_EVENT:
	return LOG_NOTICE;
	
      case RTS_FORMAT_STARTED_EVENT:
	return LOG_INFO;
	
      case RTS_FORMAT_COMPLETED_EVENT:
	return LOG_INFO;
	
      case RTS_DRIVE_VALIDATION_EVENT:
	return LOG_NOTICE;

      default:
	return LOG_NOTICE;
    }
}
